local mod = MorgDKP2:NewModule("CoreModule", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0", "AceSerializer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("MorgDKP2")
local media = LibStub("LibSharedMedia-3.0", true)
local BC = LibStub("LibBabble-Class-3.0"):GetLookupTable()
local BB = LibStub("LibBabble-Boss-3.0"):GetLookupTable()

local mdkp = MorgDKP2
mod.modName = "CoreModule"

--locals

local fmt = string.format
local sub = string.sub
local tinsert = table.insert
local tsort   = table.sort
local tremove = table.remove
local tostring = tostring 
local lower = string.lower
local find = string.find
local len = string.len
local upper = string.upper
local band = bit.band
local bor = bit.bor
local bxor = bit.bxor
local Player = UnitName("player")
local db

local defaults = { 
	profile = {
		frames = {	version = {point = "CENTER", relpoint = "CENTER", x = 0, y = 0},
				history = {point = "CENTER", relpoint = "CENTER", x = 0, y = 0},
				listdkp = {point = "CENTER", relpoint = "CENTER", x = 0, y = 0}
		},
   	},
}

local options = {
		type = "group",
		handler = mod,
		name = L["Core Module"],
		desc = L["Core of all MorgDKP2 features. \nCan not be disabled."],
		order = 5,
		args = {
			mlde = {
				type = "toggle",
				name = L["Enable ML/DE mode"],
				desc = L["If ML/DE Mode is on then no items will be recorded for the master looter or disenchanter defined by MorgDKP2."],
				get = function() return mdkp.MLmode end,
				set = function(info, v) mdkp.MLmode = v end,
				disabled = function() return not mod:IsEnabled() end,
				order = 5
			},
			mbidc = {
				type = "execute",
				name = L["Morgbid check"],
				desc = L["Determine MorgBid user base. \nAlso useful to check DKP for alts."],
				func = function() mod:MorgBidQuery() end,
				disabled = function() return not mod:IsEnabled() end,
				order = 10
			},
			hints = {
				type = "toggle",
				name = L["Enable tooltip hints"],
				desc = L["Enable tooltip hints"],
				get = function() return mdkp.db.profile.hints end,
				set = function(info, v) mdkp.db.profile.hints = v end,
				disabled = function() return not mod:IsEnabled() end,
				order = 15
			},
			display = {
				type = "group",
				handler = mod,
				name = L["Display Options"],
				desc = L["General display options for main modules."],
				order = 50,
				args = {
					font = {
		 				type = "select",
		 				dialogControl = "LSM30_Font",
		 				name = L["Font"],
		 				desc = L["Font used in tooltips and frames."],
		 				get = function() return mdkp.db.profile.media.font end,
		 				set = function(info, v) mdkp.db.profile.media.font = v mod:UpdateFrameMedia() end,
		 				values = AceGUIWidgetLSMlists.font, 
						disabled = function() return not mod:IsEnabled() end,
						order = 20
					},
					fontsize = {
						type = "range",
						name = L["Font size"],
						get = function() return mdkp.db.profile.media.fontsize end,
						set = function(info, v) mdkp.db.profile.media.fontsize = v mod:UpdateFrameMedia() end,
						min = 1, 
						max = 30, 
						step = 1,
						disabled = function() return not mod:IsEnabled() end,
						order = 25
					},
						background = {
						type = "select",
						dialogControl = "LSM30_Background",
						name = L["Background texture"],
						desc = L["Background texture of frames/tooltips."],
						get = function() return mdkp.db.profile.media.background end,
						set = function(info, v) 
								mdkp.db.profile.media.background = v
								mod:UpdateFrameMedia()
							end,
						values = AceGUIWidgetLSMlists.background, 
						disabled = function() return not mod:IsEnabled() end,
						order = 30
					},
					border = {
						type = "select",
						dialogControl = "LSM30_Border",
						name = L["Border texture"],
						desc = L["Border texture of frames/tooltips."],
						get = function() return mdkp.db.profile.media.border end,
						set = function(info, v) 
								mdkp.db.profile.media.border = v
								mod:UpdateFrameMedia()
							end,
						values = AceGUIWidgetLSMlists.border, 
						disabled = function() return not mod:IsEnabled() end,
						order = 35
					},
					backdropC = {
						type = "color",
						name = L["Background color"],
						hasAlpha = true,
						set = "SetColorOpt",
						get = "GetColorOpt",
						disabled = function() return not mod:IsEnabled() end,
						order = 40,
					},
					borderC = {
						type = "color",
						name = L["Border color"],
						hasAlpha = true,
						set = "SetColorOpt",
						get = "GetColorOpt",
						disabled = function() return not mod:IsEnabled() end,
						order = 45,
					},
					fontC = {
						type = "color",
						name = L["Font color"],
						hasAlpha = true,
						set = "SetColorOpt",
						get = "GetColorOpt",
						disabled = function() return not mod:IsEnabled() end,
						order = 50,
					},
				},
			},
        	},
}


function mod:GetOptions()
	return options
end

function mod:IsDKP()
	return nil
end

function mod:OnInitialize()
	self.db = mdkp.db:RegisterNamespace("CoreModule", defaults)
	
end

function mod:OnEnable()
	db = mdkp.db.profile
	db.moduleON["CoreModule"] = true
	mdkp.MLmode = true
	mdkp.updatedata = {}
	mdkp.lootmaster = nil 
	mdkp.AuctionStatus = {}
	mdkp.querytooltips = {}
	mdkp.roster = {}
	mdkp.oldroster = {}
	mdkp.Starttime = nil
	mdkp.origitempoints = {}
	mdkp.random = {}
	mdkp.given = nil
	mdkp.frames = {}
	mdkp.Starttime = nil
	mdkp.boelink = nil
	
	--events
	self:RegisterEvent("CHAT_MSG_LOOT", "OnLootMsg")
    	self:RegisterEvent("RAID_ROSTER_UPDATE", "UpdateRoster")
	self:RegisterEvent("GUILD_ROSTER_UPDATE", "GuildUpdate")
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "StartCombat")
	self:RegisterComm("MorgDKP2MorgBid", "OnMorgBidCommReceive")
	
	self:GetCurrentLootMethod()
	self:InitialiseRoster()
	self:IsDKPEnabled()
end

function mod:OnDisable()
	if db.alwaysloot then 
		local loot = self:GetModuleRef("Lootframe")
		if loot then loot:RegisterEvent("RAID_ROSTER_UPDATE", loot.RaidEvent) end
	end
	mdkp.frames = nil
	db.moduleON["CoreModule"] = nil
end

--called when a new DKP system is toggled on and may result in a conflict
function mod:ChangeDKPType(newtype)
	self:debug("ChangeDKP type"..newtype)
	local refname = mdkp.dkpmenuconvert[newtype]
	self:debug("ref"..refname.." "..(db.currDKP or "NIL"))
	if refname == "attemptmode" then
		db.attemptmode = not db.attemptmode
		local tracker = MorgDKP2:GetModule("Tracker", true)
		if tracker then tracker:TogglePlayerDies(db.attemptmode) end
	elseif refname == db.currDKP then
		local currDKP = MorgDKP2:GetModule(refname, true)
		db.modules[refname] = nil
		self:debug("Disabled "..refname)
		currDKP:Disable()
		db.currDKP = nil
		db.dkpstatustext = L["Not Set"]
	else
		local text = newtype
		local conflict
		if db.modules.BidWar then
			local currconflicts = db.dkpmodules.conflicts["BidWar"]
			for _, conflictname in pairs(currconflicts) do
				if conflictname == refname then 
					local bidwar = self:GetModuleRef("BidWar")
					bidwar:Disable()
					db.modules.BidWar = nil
					self:debug("Disabled BidWar due to conflict")
					conflict = true
					break
				end
			end
			if not conflict then text = text .. " & BidWar" end
		end
		if db.currDKP then
			local currDKP = MorgDKP2:GetModule(db.currDKP, true)
			db.modules[db.currDKP] = nil
			self:debug("Disabled "..db.currDKP)
			currDKP:Disable()
		end
		local newDKP = MorgDKP2:GetModule(refname, true)
		if newDKP then 
			newDKP:Enable()
			db.modules[refname] = true
			self:debug("Enabled "..refname)
		end
		db.currDKP = refname
		db.dkpstatustext = text
	end
	self:ToggleFrameUpdate("TooltipLDB")
end

function mod:UpdateFrameMedia()
	local LDB = self:GetModuleRef("LDBPlugin")
	if LDB and LDB.LDBframe then 
		self:CreateBackdrop(LDB.LDBframe)
		self:ToggleFrameUpdate("TooltipLDB", true)
	end
	local fubar = self:GetModuleRef("FubarPlugin")
	if fubar and fubar.Fubarframe then 
		self:CreateBackdrop(fubar.Fubarframe)
		self:ToggleFrameUpdate("TooltipFubar", true)
	end
	local lootframe = self:GetModuleRef("Lootframe")
	if lootframe and lootframe.lootframe then 
		self:CreateBackdrop(lootframe.lootframe)
		self:CreateBackdrop(lootframe.itemframe, true)
		self:CreateBackdrop(lootframe.rollframe, true)
		self:CreateBackdrop(lootframe.rollsummaryframe, true)
		lootframe.font = media:Fetch("font", db.media.font)
		lootframe.itemframe:Hide()
		lootframe.itemframe = lootframe:CreateItemsFrame()
		lootframe:UpdateLootFrame(nil, nil, nil, nil, true, true)
	end
	for name, data in pairs(mdkp.frames) do
		self:CreateBackdrop(data)
		self:ToggleFrameUpdate(name)
	end
end

function mod:CreateBackdrop(f, skip)
	local backdrop = f:GetBackdrop()
	local font = media:Fetch("font", db.media.font)
	if not backdrop then
		backdrop = {
			insets = {left = 3, right = 3, top = 3, bottom = 3},
			tile = false, 
			tileSize = 16, 
			edgeSize = 8,
		}
	end
	backdrop.edgeFile = media:Fetch("border", db.media.border)
	backdrop.bgFile = media:Fetch("background", db.media.background)
	f:SetBackdrop(backdrop)
	f:SetBackdropColor(unpack(db.media.backdropC))
   	f:SetBackdropBorderColor(unpack(db.media.borderC))
   	if f.title and not skip then 
   		f.title:SetTextColor(unpack(db.media.borderC)) 
   		f.title:SetFont(font, db.media.fontsize + 1)
   	end
   	if f.hint then
   		f.hint:SetFont(font, db.media.fontsize) 
   	end
end

function mod:SetColorOpt(info, r, g, b, a)
	local db = mdkp.db.profile.media
	local opt = info[#info]
	db[opt][1] = r
	db[opt][2] = g
	db[opt][3] = b
	db[opt][4] = a
	self:UpdateFrameMedia()
end

function mod:GetColorOpt(info)
	local db = mdkp.db.profile.media
	local opt = info[#info]
	return unpack(db[opt])
end

function mod:OnLootMsg(_, lootmsg)
	if not db.InRaid then return end
	--self:debug("Loot message")
	local sPlayer, sLink
	_, _, sPlayer, sLink = find(lootmsg, L["([^%s]+) receives loot: (.+)%."])
  	if not sPlayer or not sLink then
		_, _, sLink = find(lootmsg, L["You receive loot: (.+)%."])
    		sPlayer = Player
	end
	
	-- Make sure there is a link
	if not mdkp.roster[sPlayer] then self:debug("Player not in raid "..sPlayer) return end
	if sLink and sPlayer then
		self:RecordLoot(sLink, sPlayer)
	end
end

function mod:InitialiseRoster()
	self:debug("Initialising roster..")
	local totalmembers = GetNumRaidMembers()
	if totalmembers == 0 then
		db.raidmembers = {Player}
		mdkp.roster[Player] = true
	else
		db.raidmembers = {}
		for i = 1, totalmembers do
			local name = UnitName("raid"..i)
			mdkp.roster[name] = true
			tinsert(db.raidmembers, name)
		end
	end
	mdkp.oldroster = mdkp.roster
	tsort(db.raidmembers)
end

function mod:UpdateRoster()
	--self:debug("Updating roster..")
	
	if not UnitInRaid("player") and not mdkp.querytooltips["listdkp"] then return end 
	mdkp.oldroster = mdkp.roster
	mdkp.roster = {}
	local totalmembers = GetNumRaidMembers()
	for i = 1, totalmembers do
		local name = UnitName("raid"..i)
		mdkp.roster[name] = true
	end
	--new member?
	for name in pairs(mdkp.roster) do
		if not mdkp.oldroster[name] then
			self:AddMember(name)
		end
	end
	--member left?
	for name in pairs(mdkp.oldroster) do
		if not mdkp.roster[name] and name ~= Player and not UnitCanAttack("player", name) then
			self:RemoveMember(name)
		end
	end
	if mdkp.querytooltips["listdkp"] then 
		mdkp.querytooltips["listdkp"] = self:Onlinemembers(mdkp.querytooltips["listdkp"]) 
		self:ToggleFrameUpdate("listdkp")
	end
	--self:RequestLeader()
	tsort(db.raidmembers)
	self:GetCurrentLootMethod()
end

function mod:AddMember(name)
	self:debug("Add member.."..name)
	local raidnum = db.raidnum
	local ttime = date("%m/%d/%y %H:%M:%S")
	local race = UnitRace(name)
    	local level = UnitLevel(name)
	local guild = GetGuildInfo(name)
	local classtrue, englishclass = UnitClass(name)
	local tracker = self:GetModuleRef("Tracker")
	if raidnum > 0 then
		if not db.raidlog[raidnum].playerinfo then db.raidlog[raidnum].playerinfo = { } end
		if not db.raidlog[raidnum].playerinfo[name] then db.raidlog[raidnum].playerinfo[name] = { } end
		db.raidlog[raidnum].playerinfo[name].class = classtrue 
		db.raidlog[raidnum].playerinfo[name].race = race 
		db.raidlog[raidnum].playerinfo[name].level = level 
		db.raidlog[raidnum].playerinfo[name].guild = guild 
		if not db.raidlog[raidnum].join then db.raidlog[raidnum].join = { } end
		if not db.raidlog[raidnum].join[name] then 
			db.raidlog[raidnum].join[name] = ttime 
			if db.raidlog[raidnum].leave then
				if db.raidlog[raidnum].leave[name] then db.raidlog[raidnum].leave[name] = nil end
			end
			self:out(fmt(L["%s joined the raid at %s."], name, ttime))
			if tracker then tracker:StartAttEvent() end
		end
	end
	self:PlayerinDB(name, englishclass)
	
	--update listdkp
	if mdkp.querytooltips["listdkp"] then 
		local found = nil
		for _, data in pairs(mdkp.querytooltips["listdkp"]) do
			if data[1].n == name then 
				found = true
				break
			end
		end
		if not found then
			local POOL = db.raid
			local namedata, pointsdata = self:GetDKPInfo(name, POOL)
			local status = mdkp.guildstatustext[L["INRAID"]]
			tinsert(mdkp.querytooltips["listdkp"], {namedata, status, pointsdata}) 
		end
	end
	
	--update waitlist
	local wait = self:GetModuleRef("Waitlist")
	if wait then wait:AtMemberAdded(name) end
	
	tinsert(db.raidmembers, name)
end

function mod:RemoveMember(name)
	self:debug("Remove member.."..name)
	local raidnum = db.raidnum
	local ttime = date("%m/%d/%y %H:%M:%S")
	if raidnum > 0 then
		if not db.raidlog[raidnum].leave then db.raidlog[raidnum].leave = { } end
		db.raidlog[raidnum].leave[name] = ttime
		self:out(fmt(L["%s left the raid at %s."], name, ttime))
	end
	for num, listname in pairs(db.raidmembers) do
		if name == listname then
			tremove(db.raidmembers, num)
			break
		end
	end
end

function mod:GuildUpdate()
	--self:debug("Guild update..")
	if mdkp.querytooltips["listdkp"] then
		mdkp.querytooltips["listdkp"] = self:Onlinemembers(mdkp.querytooltips["listdkp"])
		self:ToggleFrameUpdate("listdkp")
	end
end

function mod:StartCombat()
	if not db.InRaid then return end
	self:debug("Enter combat")
	mdkp.Starttime = date("%m/%d/%y %H:%M:%S")
end

function mod:Scan()
	local SLOTS = {
		"HeadSlot",
		"NeckSlot",
		"ShoulderSlot",
		"BackSlot",
		"ChestSlot",
		"WristSlot",
		"HandsSlot",
		"WaistSlot",
		"LegsSlot",
		"FeetSlot",
		"Finger0Slot",
		"Finger1Slot",
		"Trinket0Slot",
		"Trinket1Slot",
		"MainHandSlot",
		"SecondaryHandSlot",
		"RangedSlot"
	}
	
	for i = 1, GetNumRaidMembers() do
		local name = UnitName("raid"..i)
		local _, class = UnitClass("raid"..i)
		self:PlayerinDB(name, class)
		if not db.info[name].items then db.info[name].items = { } end
		if CheckInteractDistance(name, 1) then
			NotifyInspect(name)
			for _,slot in pairs(SLOTS) do
				local itemLink = GetInventoryItemLink(name, GetInventorySlotInfo(slot))
				if itemLink then
					local _, _, itemCode = find(itemLink, "(%d+):")
					itemCode = tonumber(itemCode)
					if not self:ItemRecorded(itemCode, name) then
						tinsert(db.info[name].items, itemCode)
						self:debug("Added ".. name .. " slot: ".. slot ..";item ".. itemLink ..";ID: " .. itemCode)
					end
				else self:debug(name.."'s slot: "..slot.."; no item")
				end
			end
			self:debug("Scanned "..name) 
		end
	end
end

function mod:ItemRecorded(item, name)
	for _, ID in pairs(db.info[name].items) do
		if item == ID then
			return true
		end
	end
	return false
end

function mod:ItemQuery(id, link)
	if not self:ValidRaid() then 
		self:out(L["You must start a raid before you can start a query."]) 
		return false 
	end
	
	--scan current members for gear
	self:Scan()
	
	--who needs the query?
	needquery = self:HasItem(id)
	if not needquery or #needquery == 0 then 
		self:out(L["Either no classes are selected or no members need this item."]) 
		return false
	end
	
	if mdkp.AuctionStatus[id] then 
		mdkp.AuctionStatus[id] = mdkp.AuctionStatus[id] + 1
		return false 
	else mdkp.AuctionStatus[id] = 1
	end
	
	local itempoints = db.items[id].points or 0
	mdkp.origitempoints[id] = itempoints
	local POOL = self:CheckPool(id)
	local points
	mdkp.querytooltips[id] = {}
  		
	--module stuff
	self:IsDKPEnabled()
	local bidwar = self:GetModuleRef("BidWar")
	if bidwar then bidwar:AtStartQuery(id) end 
	local whisper = self:GetModuleRef("Whisper")
	local currDKP = self:GetModuleRef(db.currDKP)
	local lootframe = self:GetModuleRef("Lootframe")
	local random = self:GetModuleRef("Random")
	
	--send queries & create query table
	for _, data in pairs(needquery) do
		self:PointsPoolExists(data.n, POOL)
		local dkppoints = currDKP:GetPlayerPoints(data.n, POOL)
		if mdkp.random[id] or random then dkppoints = L["NONE"] end
		if not dkppoints then dkppoints = db.info[data.n].points end
		
		if bidwar then bidwar:SendMessage(id, itempoints, data.n, dkppoints)
		else self:SendCommMessage("MorgBid2", self:Serialize("QUERY", id, itempoints, data.n, dkppoints), "RAID")
		end
		if whisper then whisper:SendStartMessage(id, link, itempoints, data.n) end
		if dkppoints == "NONE" or bidwar then dkppoints = 0 end
		local points = {n = dkppoints, r = 0, g = 1, b = 0}
		if db.info[data.n].raidloot == 1 then points.r = 1 end
		tinsert(mdkp.querytooltips[id], {data, mdkp.statustext["PENDING"], points, data.p})
	end
	self:debug(fmt("Query sent %s for %s DKP", id, itempoints))
	
	--create output message for classes
	local mainstring = L["Main: "]
	local altstring = L["Offspec: "]
	local isaltstring = nil
	if db.items[id].classes == 2047 then mainstring = mainstring .. L["ALL"] ..", "
	else
		for bitclass, bit in pairs(mdkp.Bits) do
			if bitclass ~= L["ALL"] then
				if band(db.items[id].classes, bit) == bit then
					mainstring = mainstring .. bitclass .. ", "
				end
				if band(db.items[id].altclasses, bit) == bit then
					altstring = altstring .. bitclass .. ", "
					isaltstring = true
				end
			end
		end
	end
	if bidwar then bidwar:StartOutput(link, id, itempoints) 
	else
		SendChatMessage(fmt(L["Now querying for %s  ID = %s  DKP = %s"], link, id, itempoints), "RAID")
		mainstring = sub(mainstring, 1, -3)
		SendChatMessage(mainstring, "RAID")
		if isaltstring then
			altstring = sub(altstring, 1, -3)
			SendChatMessage(altstring, "RAID")
		end
	end
  	lootframe:SortItemTT(id)
  	return true
end

function mod:HasItem(id)
	if not db.items[id] then return end
	local classes = db.items[id].classes
	local altclasses = db.items[id].altclasses
	local _, _, _, _, _, _, _, _, iEquipLoc, _ = GetItemInfo(id)
	local needquery = {}
	local neededclasses = {}
	local needed = nil
	
	--set up classlist
	for bitclass, bit in pairs(mdkp.Bits) do
		if bitclass ~= L["ALL"] then
			if band(classes, bit) == bit then
				neededclasses[bitclass] = 1	--1st priority
				needed = true
			elseif band(altclasses, bit) == bit then
				neededclasses[bitclass] = 2	--2nd priority
				needed = true
			end
		end
	end
	if not needed then return end
					
	--check roster for needy classes
	local hasitem
	for pname in pairs(mdkp.roster) do
		local pdata = self:GetClasscolors(pname)
		if neededclasses[pdata.c] then
			if db.info[pname] and db.info[pname].items then
				local items = db.info[pname].items
				hasitem = nil
				for _, hasID in pairs(items) do
					if hasID == id and (id ~= 19147 or id ~= 19140 or id ~= 17071 or id ~= 32527) then
						hasitem = 1
					end
				end
			end
			if not hasitem then
				pdata.p = neededclasses[pdata.c]	--insert priority data
				tinsert(needquery, pdata)
			end
		end
	end
	return needquery
end

function mod:MorgBidQuery()
	self:debug("Morgbid check")
	--if not self:ValidRaid() then return end
	local POOL = db.raid
	
	--modules
	self:IsDKPEnabled()
	local currDKP = self:GetModuleRef(db.currDKP)
	
	--populate query table
	mdkp.querytooltips["version"] = {}
  	for name in pairs(mdkp.roster) do
  		self:debug(name or "NIL")
  		self:PointsPoolExists(name, POOL)
  		local points = currDKP:GetPlayerPoints(name, POOL)
  		local namedata = self:GetClasscolors(name)
  		local pointsdata = {n = points, r = 0, g = 1, b = 0}
  		if db.info[name].raidloot == 1 then pointsdata.r = 1 end
  		tinsert(mdkp.querytooltips["version"], {namedata,  mdkp.statustext[L["NONE"]], pointsdata})
  	end
  	
  	--fetch a free frame & create popup
  	local hint = L["|c000070ddClick:|r Send tells.\n|c000070ddR-Click:|r Close tablet."]
   	local title = L["MorgBid2 Version Query"]
   	local f = self:GetFrame("version", hint, title)
  	self:ToggleFrameUpdate("version")
  	f:Show()
  	self:SendCommMessage("MorgBid2", self:Serialize("CHECK", "A"), "RAID")
end

function mod:WhisperUsers()
	self:debug("Whispering out of date morgbid users")
	for _, data in pairs(mdkp.querytooltips["version"]) do
		if data[2].n == L["NONE"] or data[2].n < mdkp.morgbid2version then SendChatMessage(L["Please update your MorgBid2.  http://www.wowace.com/projects/morg-bid2/ or check in the vent chat window.  Thank-you:)"], "WHISPER", nil, data[1].n) end
	end
end

function mod:ListDKP(listtext)
	if not listtext then return end
	self:debug("ListDKP")
	dkplist = self:FillDKPTable(nil, nil, listtext, nil, true)
	
	--update guild members
	mdkp.querytooltips["listdkp"] = self:Onlinemembers(dkplist)
	
	--fetch a free frame & create popup
  	local hint = L["|c000070ddL-Click:|r Invite.\n|c000070ddCTRL-Click:|r Delete Character.\n|c000070ddALT-Click:|r Add to waitlist.\n|c000070ddSHIFT-Click:|r Remove from waitlist.\n|c000070ddR-Click:|r Close tablet."]
   	local title = L["List DKP"]
   	local f = self:GetFrame("listdkp", hint, title)
  	f:Show()
  	self:ToggleFrameUpdate("listdkp")
end

function mod:OnListDKPClick(name, status)
	if IsShiftKeyDown() then
		local wait = self:GetModuleRef("Waitlist")
		if wait then wait:SubWait(name) end
	elseif IsControlKeyDown() then
		db.info[name] = nil
		for num, v in pairs(mdkp.querytooltips["listdkp"]) do
			if v[1].n == name then
				tremove(mdkp.querytooltips["listdkp"], num)
				break
			end
		end
		self:out(fmt(L["Deleting %s from MorgDKP2 database."], name))
		self:ToggleFrameUpdate("listdkp")
	elseif IsAltKeyDown() then
		local wait = self:GetModuleRef("Waitlist")
		if wait then wait:AddWait(name) end
	else
		if status == L["WAIT"] or status == L["xWAITx"] then
			local wait = self:GetModuleRef("Waitlist")
			if wait then name = wait:InformOnlineAlias(name, self:GetOnlineGuildMembers()) end
		end
		InviteUnit(name)
	end
end

function mod:ShowAliasTip(member)
	local POOL = db.raid
	local alts = {}
	for alt, main in pairs(db.eqDKP[POOL].Aliases) do
		if main == member then tinsert(alts, alt) end
	end
	if alts and #alts > 0 then
		GameTooltip:AddLine("Aliases:", 1, 1, 1)
		for _, alt in pairs(alts) do
			local name = self:GetClasscolors(alt)
			GameTooltip:AddLine(alt, name.r, name.g, name.b)
		end
		GameTooltip:Show()
	end
  		
end

function mod:Onlinemembers(dkplist)
	local guildmembers = self:GetOnlineGuildMembers()
	local status
	local POOL = db.raid
	local done = {}
	local wait = self:GetModuleRef("Waitlist")
	for num, data in pairs(dkplist) do
		data[2] = self:GetStatus(data[1].n, guildmembers, wait)
		guildmembers[data[1].n] = nil
	end
	for name, data in pairs(guildmembers) do
		if data.o and db.info[name] and not done[name] then 
			local namedata, pointsdata = self:GetDKPInfo(name, POOL)
			local status = mdkp.guildstatustext[L["ONLINE"]]
			tinsert(dkplist, {namedata, status, pointsdata}) 
		end
	end
	return dkplist
end

function mod:GetStatus(name, guildmembers, wait)
	local inraid = UnitInRaid(name)
	local guilddata = guildmembers[name]
	if wait then isonwait = wait:IsOnWaitlist(name) end
	if guilddata then 
		if guilddata.o then
			if inraid then
				if guilddata.s == "<AFK>" then status = mdkp.guildstatustext["<AFK>"]
				else status = mdkp.guildstatustext[L["INRAID"]] 
				end
			elseif isonwait then status = mdkp.guildstatustext[L["WAIT"]]
			else status = mdkp.guildstatustext[L["ONLINE"]]
			end
		elseif inraid then self:debug(name.."xINRAIDx") status = mdkp.guildstatustext[L["xINRAIDx"]] 
		elseif isonwait then status = mdkp.guildstatustext[L["xWAITx"]]
		else status = mdkp.guildstatustext[L["OFFLINE"]]
		end
	elseif inraid then status = mdkp.guildstatustext[L["INRAID"]] 
	elseif isonwait then status = mdkp.guildstatustext[L["WAIT"]]
	else status = mdkp.guildstatustext[L["xGUILDx"]] 
	end
	return status
end

function mod:GetOnlineGuildMembers()
	local numGuildMembers = GetNumGuildMembers(true)
	local Onlinemembers = { }
	for i = 1, numGuildMembers do
		local name, _, _, _, _, _, _, _, online, status = GetGuildRosterInfo(i)
		Onlinemembers[name] = {o = online, s = status}
	end
	return Onlinemembers
end

function mod:FillDKPTable(sender, senderclass, txt, raided, showstatus)
	local POOL = db.raid
	dkplist = { }
	txt = upper(txt)
	
	--add for ALL/add sender if needed
	if find(txt, "ALL") then
		txt = "DKPLIST DRUID HUNTER MAGE PALADIN PRIEST ROGUE SHAMAN WARLOCK WARRIOR DEATHKNIGHT"
	elseif senderclass and not find(txt, upper(senderclass)) then
		txt = txt .. " ".. upper(senderclass)
	end
	
	local args = {self:Explode(txt, " ")}
	
	--is pool in txt?
	for _, name in pairs(db.raidlist) do
		if upper(name) == args[2] then
			POOL = name
		end
	end
	
	--modules
	self:IsDKPEnabled()
	
	--format table for output
	local guildmembers = self:GetOnlineGuildMembers()
	local wait = self:GetModuleRef("Waitlist")
	numargs = #args
	for i = 2, numargs do
		local class = args[i]
		if raided then
			for pname in pairs(mdkp.roster) do
				local pclass = db.info[pname].class
				if pclass == class then
					local namedata, pointsdata = self:GetDKPInfo(pname, POOL)
  					tinsert(dkplist, {namedata, pointsdata}) 
				end
			end
		else
			for pname, data in pairs(db.info) do
				if data.class and data.class == class then
					local status = self:GetStatus(pname, guildmembers, wait) 
					if not db.eqDKP[POOL].Aliases[pname] or (db.eqDKP[POOL].Aliases[pname] and (status == L["ONLINE"] or status == "<AFK>" or status == L["INRAID"] or status == L["xINRAIDx"] or status == L["WAIT"] or status == L["xWAITx"])) then
						local namedata, pointsdata = self:GetDKPInfo(pname, POOL)
  						if showstatus then tinsert(dkplist, {namedata, status, pointsdata}) 
  						else tinsert(dkplist, {namedata, pointsdata}) end
  					end
  				end
			end
		end
	end
	return dkplist, POOL
end

function mod:OnMorgBidCommReceive(prefix, message, distribution, sender)
	--self:debug("Addon message..."..prefix..message..distribution..sender)
	local result, command, item, points = self:Deserialize(message)
	if command == L["PASS"] or command == L["TAKE"] or command == L["NEED"] then
		self:debug("From "..sender..": "..command)
		if not mdkp.querytooltips[item] then
			self:debug("No query for "..item)
			return
		end
		local bidwar = self:GetModuleRef("BidWar")
		local rand = self:GetModuleRef("Random")
		for i,v in ipairs(mdkp.querytooltips[item]) do
			if v[1].n == sender then
				v[2] = mdkp.statustext[command]
				if command ~= L["PASS"] then
					if bidwar then v[3].n = bidwar:OnBidMessage(points, sender, item) 
					elseif rand or mdkp.random[item] then v[3].n = math.random(100)
					end
					break
				end
			end
		end
		self:ToggleFrameUpdate(item)
	elseif command == "REPLY" then
		if not mdkp.querytooltips["version"] then
			self:debug("No version query exists.") 
			return
		end
		for i,v in ipairs(mdkp.querytooltips["version"]) do
			if v[1].n == sender then 
				local version = {n = points, r = 0, g = 1, b = 0}
				if points == L["NONE"] then
					version.r, version.g = 1, 0
				elseif points and points < mdkp.morgbid2version then version.r = 1 end
				v[2] = version
				break
			end
		end
		self:ToggleFrameUpdate("version")
	end
end

function mod:GetFrame(name, hint, title)
	if not mdkp.frames[name] then
		local ondragstart = function(f) f:StartMoving() end
		local ondragstop =  function(f)
   						f:StopMovingOrSizing()
   						local point, _, relativePoint, x, y = f:GetPoint()
						self.db.profile.frames[name] = {point = point, relpoint = relativePoint, x = x, y = y}
   					end
   		local onmouseup = function(f, button) if button == "RightButton" then mdkp.querytooltips[f.datacheck] = nil f:Hide() end end 
   		local onupdate = function(f) self:FrameUpdate(f) end
   		mdkp.frames[name] = self:Createframe(nil, nil, nil, onupdate, onmouseup, hint, name, nil, nil, title, ondragstart, ondragstop, true)
		
		local f = mdkp.frames[name]
		local framepos = self.db.profile.frames[name]
		local point, relpoint, x, y = framepos.point, framepos.relpoint, framepos.x, framepos.y
		f:SetPoint(point, UIParent, relpoint, x, y)
	
		f.close = CreateFrame("Button", nil, f)
   		f.close:SetScript("OnClick", function(f) mdkp.querytooltips[f:GetParent().datacheck] = nil f:GetParent():Hide() end)
   		f.close:SetWidth(20)
   		f.close:SetHeight(20)
   		f.close:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up")
   		f.close:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down")
   		f.close:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight")
   		f.close:ClearAllPoints()
   		f.close:SetPoint("TOPRIGHT", f, "TOPRIGHT", 0, 0)
   		f.close:SetHitRectInsets(5, 5, 5, 5)
   		f.close:Show()
   	end
   	return mdkp.frames[name]
end

function mod:IncomingWhispers(command, sender, actions)
	local raided = UnitInRaid(sender)
	
	--player class
	local pdata = self:GetClasscolors(sender)
	local pclass = pdata.c
	
	--actions
	if command == "MBID" then 
		--waitlist??
		if find(actions, "WAIT") then 
			local wait = self:GetModuleRef("Waitlist")
			if wait then wait:OnWaitWhisper(sender, actions) end
			return 
		end
		
		--format correct?
		local args = {}
		local _,_, item = find(actions, "ITEM:(%d+):")
		if item then
			args[1] = sub(actions, 1, 4)
			args[2] = item
			local sstart, send = find(actions, "]")
			if sstart then args[3] = sub(actions, sstart + 6) end
		else 
			args = {self:Explode(actions, " ")}
			item = args[2]
		end
		item = tonumber(item)
		if #args < 3 or not mdkp.querytooltips[item] then SendChatMessage(L["Sorry your response format is incorrect or no item with that ID is up for bidding.  Format = mbid ID [need/take/pass/bid value]"], "WHISPER", nil, sender) return end
	
		--item in auction?
		if not mdkp.AuctionStatus[item] then
			SendChatMessage(L["This item is not currently up for bid."], "WHISPER", nil, sender)
			return
		end
	
		if args[3] == L["PASS"] or args[3] == L["TAKE"] or args[3] == L["NEED"] then
			for i,v in pairs(mdkp.querytooltips[item]) do
				if v[1].n == sender then
					v[2] = mdkp.statustext[args[3]]
					SendChatMessage(fmt(L["Your response has been accepted: %s"], args[3]), "WHISPER", nil, sender)
					local rand = self:GetModuleRef("Random")
					if (rand or mdkp.random[item]) and args[3] ~= L["PASS"] then v[3].n = math.random(100) end
				end
			end
		else
			--bidwar??
			local bidwar = self:GetModuleRef("BidWar")
			if bidwar then bidwar:OnBidWhisper(sender, item, args[3]) end
		end
		self:ToggleFrameUpdate(item)
		return
	end
	--output pools if toggle
	if command == "DKPL" and find(actions, "POOLS") then 
		self:ShowPools(sender)
		return
	end
	
	--send dkplist results
	dkplist, pool = self:FillDKPTable(sender, pclass, actions, raided)
	sort(dkplist, function(a,b) return a[2].n > b[2].n end)
	local distribution
	if IsInGuild() then distribution = "GUILD" end
	if UnitInRaid(sender) then distribution = "RAID" end
	if distribution then 
		self:SendCommMessage("MorgBid2", self:Serialize("INITIATE", sender), distribution)
		self:SendCommMessage("MorgBid2", self:Serialize("DKPL", sender, dkplist, pool), distribution)
	else self:debug("Not in guild!  Can't send addon message.") end
end

function mod:ShowPools(sender)
	SendChatMessage(L["Currently loaded DKP pools..."], "WHISPER", nil, sender)
	local counter = 1
	for _, name in pairs(db.raidlist) do
		SendChatMessage(counter .. ". " .. name, "WHISPER", nil, sender)
		counter = counter + 1
	end
	SendChatMessage(L["Please whisper: dkplist [poolname] [class] [class] .... [all]"], "WHISPER", nil, sender)
end

function mod:AddAlias(alt, main, quiet)
	self:debug("addalias"..alt.." of "..main)
	local POOL = db.raid
	mainpoints = 0
	
	--return if no main
	if db.info[main] then mainpoints = db.info[main][POOL].points
	else
		if not quiet then self:out(fmt(L["Main character %s does not exist.  No changes made."], main)) end
		return
	end
	
	--return if alias exists
	if db.eqDKP[POOL].Aliases and db.eqDKP[POOL].Aliases[alt] then
		if not quiet then self:out(fmt(L["Alias %s already exihists.  No changes made."], alt)) end
		return
	end
	
	--add alias
	self:out(fmt(L["Transeferred %s DKP to %s from %s"], mainpoints, alt, main))
	self:PlayerinDB(alt)
	db.info[alt][POOL]  =  {points = mainpoints,
				earned = db.info[main][POOL].earned,
				spent = db.info[main][POOL].spent}
	db.info[alt].raidloot = 0
	if not self:DoesAliasExist(main, alt, POOL) then tinsert(db.info[main].aliases, {raid = POOL, alt = alt}) end
	local update = self:GetServerTime()
	db.info[main].lastupdate = update
	db.info[alt].lastupdate = update
	if not db.raidlog.aliases[POOL] then db.raidlog.aliases[POOL] = { } end
	tinsert(db.raidlog.aliases[POOL], {alt = alt,
						main = main,
						eqdkp = db.eqDKP[POOL].eqDKPsite,
						action = "ADD" })
	db.eqDKP[POOL].Aliases[alt] = main
end

function mod:DoesAliasExist(main, alt, pool)
	local match = nil
	for _, aliasdata in pairs(db.info[main]["aliases"]) do
		if aliasdata.alt == alt and aliasdata.raid == pool then
			match = true
			break
		end
	end
	return match
end

function mod:GiveBOE(name)
	self:debug("Give BOE to "..name)
	local itemlink = mdkp.boelink
	mod:RecordLoot(itemlink, name)
end

function mod:RecordLoot(itemlink, player)
	if db.raidnum < 1 then self:out(L["No current raid exists."]) return end
	local _,_, itemId = find(itemlink, "item:(%d+):")
	itemId = tonumber(itemId)
	
	--is this item ignored?
	for _, ignorelink in pairs(db.ignore) do
		local _,_, ignoreid = find(ignorelink, "item:(%d+):")
		ignoreid = tonumber(ignoreid)
		if ignoreid == itemId then self:debug("Item Ignored") return end
	end
	
	--get item info
	local sName, iLink, iQuality, _, _, class, subclass, _, iEquipLoc, iconGIF = GetItemInfo(itemId)
	
	--return if qual < threshold
	if iQuality < db.quality then return end
	
	--determine what mob dropped item
	local lootTime = date("%m/%d/%y %H:%M:%S")
	local lastboss = db.raidlog.lastboss
	local lastmob = db.raidlog.lastmob
	local BossMob
	local disenchanted
	if db.lootbuffer and db.lootbuffer[sName] ~= nil then
		BossMob = db.lootbuffer[sName]
	else
		if lastboss == lastmob then BossMob = lastboss
		else BossMob = lastmob end
	end
	if BossMob == BB["Gathios the Shatterer"] or BossMob == BB["High Nethermancer Zerevor"] or BossMob == BB["Lady Malande"] or BossMob == BB["Veras Darkshadow"] then
		BossMob = BB["Illidari Council"]
	end
	local match = nil
	for _, bossname in pairs(mdkp.ZoneBosses[db.menuzone]) do
  		if BossMob == bossname then
  			match = true
  			break
  		end
  	end
  	if not BossMob or db.OneEvent or not match then BossMob = L["Trash mob"] end
	if mdkp.MLmode and (player == db.mlooter or player == db.disenchanter) and not mdkp.given then 
		disenchanted = true
		BossMob = L["Bank"]
	else mdkp.given = nil end
	
	--modules
	self:IsDKPEnabled()
	local currDKP = self:GetModuleRef(db.currDKP)
	local bidwar = self:GetModuleRef("BidWar")
	
	--determine item value etc
	self:IteminDB(itemId)
	local POOL = self:CheckPool(itemId)
	self:PointsPoolExists(player, POOL)
	local points = currDKP:GetPlayerPoints(player, POOL)
	local itempoints = currDKP:GetItemValue(itemId, player, POOL)
	local Zone = GetRealZoneText()
	
	--insert item data into raidlog
	tinsert(db.raidlog[db.raidnum].loot,
					{
						["ItemName"] = sName,
						["ItemLink"] = iLink,
						["Quality"] = iQuality,
						["ID"] = itemId,
						["Icon"] = iconGIF,
						["Class"] = class,
						["SubClass"] = subclass,
						["Player"] = player,
						["Costs"] = itempoints,
						["Time"] = lootTime,
						["Zone"] = Zone,
						["Boss"] = BossMob,
						["eqdkp"] = db.eqDKP[POOL].eqDKPsite,
						["prefix"] = db.eqDKP[POOL].prefix
					}
				)
	
	if not disenchanted then
		--deduct points from player etc
		currDKP:ItemRecorded(itemId, player, POOL, points, itempoints, iLink, disenchanted)
		if bidwar then bidwar:ItemRecorded(itemId) end
	
		--add item into player DB
		if iEquipLoc ~= "" then
			if not db.info[player].items then db.info[player].items = { } end
			if not self:ItemRecorded(itemId, player) then
				tinsert(db.info[player].items, itemId)
			end
		end
	end
	self:ToggleFrameUpdate("TooltipLDB")
end
